
#include <stdio.h>
#include <string.h>

#define MAX_PALETTE     256 // entries
#define MAX_SCANLINE 131072 // bytes
#define MAX_TILES      1024 // horizontal

#define ERR_OPEN  "Error opening file"
#define ERR_READ  "Error reading file"
#define ERR_WRITE "Error writing file"

#pragma pack(push,1)
typedef struct {
	__int16 bm;
	__int32 fsize,reserved,dataofs,headsize,xsize,ysize;
	__int16 planes,bpp;
	__int32 compress,mapsize,xres,yres,palsize,palimp;
} BMPhead;
#pragma pack(pop)

int roundup_bits(int n) {
	return ((n+31)/32)*4;
}

void bitcopy(unsigned char *dst,const unsigned char *src,int dst_offset,int src_offset,int len) {
  int dst_int = dst_offset / 8 , src_int = src_offset / 8;
  int dst_mod = dst_offset % 8 , src_mod = src_offset % 8;
  int len_mid = (len + dst_mod) / 8 - 1;
  int shift = (src_mod - dst_mod) % 8;
  if (shift<0) { shift += 8; src_int--; }
  int inv_shift = 8 - shift;

  unsigned char mask1 = (0xFF>>dst_mod);
  unsigned char mask2 = (0xFF<<8-(dst_offset+len)%8);
  unsigned char byte1 = mask1 & ((src[src_int]<<shift) | (src[src_int+1]>>inv_shift));
  unsigned char byte2 = mask2 & ((src[src_int+len_mid+1]<<shift) | (src[src_int+len_mid+2]>>inv_shift));

  if (len_mid>=0) {
    dst[dst_int] = (~mask1 & dst[dst_int]) | byte1;
    for (int i=1; i<=len_mid; i++) {
      dst[dst_int+i] = (src[src_int+i]<<shift) | (src[src_int+i+1]>>inv_shift);
    }
    dst[dst_int+len_mid+1] = (~mask2 & dst[dst_int+len_mid+1]) | byte2;
  }
  else {
    dst[dst_int] = ~(mask1 & mask2) & dst[dst_int] | (byte1 & byte2);
  }
}

bool split_one(const char *name_in, const char *name_out, int tilesizex, int tilesizey, int digits, int overlap) {
	BMPhead head_in,head_out;
	unsigned char palette[4*MAX_PALETTE];
	unsigned char buffer_in[MAX_SCANLINE];
	unsigned char buffer_out[MAX_SCANLINE];
	int read,written;
	FILE* fp_in = NULL;
	FILE* fp_out1[MAX_TILES]; memset(&fp_out1,0,MAX_TILES*sizeof(FILE*));
	FILE* fp_out2[MAX_TILES]; memset(&fp_out2,0,MAX_TILES*sizeof(FILE*));

	try {
		fp_in = fopen(name_in,"rb");
		if (!fp_in) throw ERR_OPEN;

		printf("Reading %s ...\n",name_in);

		read = fread(&head_in,1,sizeof(head_in),fp_in);
		if (read<sizeof(head_in)) throw ERR_READ;

		int palsize_bytes = head_in.dataofs-sizeof(head_in);

		read = fread(&palette,1,palsize_bytes,fp_in);
		if (read<palsize_bytes) throw ERR_READ;

		head_out=head_in;
		head_out.xsize = tilesizex;
		int tilesx = (head_in.xsize-overlap) / (tilesizex-overlap);
		head_out.ysize = tilesizey;
		int tilesy = (head_in.ysize-overlap) / (tilesizey-overlap);
		
		if ((tilesx>   10 || tilesy>   10) && digits<2) digits = 2;
		if ((tilesx>  100 || tilesy>  100) && digits<3) digits = 3;
		if ((tilesx> 1000 || tilesy> 1000) && digits<4) digits = 4;
		if ((tilesx>10000 || tilesy>10000) && digits<5) digits = 5;

		int scanline_in_bits  = head_in.xsize*head_in.bpp;
		int scanline_out_bits = head_out.xsize*head_out.bpp;
    int scanline_in_step_bits = (tilesizex-overlap)*head_in.bpp;
		int rounded_scanline_in  = roundup_bits(scanline_in_bits);
		int rounded_scanline_out = roundup_bits(scanline_out_bits);
		head_out.fsize = head_out.dataofs+tilesizey*rounded_scanline_out;

		// SKIP DISCARDED BOTTOM LINES
		for (int i=0; i<(head_in.ysize-tilesizey*tilesy+overlap*(tilesy-1)); i++) {
			read = fread(&buffer_in,1,rounded_scanline_in,fp_in);
			if (read<rounded_scanline_in) throw ERR_READ;
		}

		// SPLIT OTHER LINES
		for (int pixy=0; pixy<tilesizey*tilesy-overlap*(tilesy-1); pixy++) {
			/// OPEN FILES
			if (pixy%(tilesizey-overlap)==0) {
				memcpy(&fp_out2,&fp_out1,MAX_TILES*sizeof(FILE*));
				memset(&fp_out1,0,MAX_TILES*sizeof(FILE*));
				if (pixy!=tilesy*(tilesizey-overlap)) {
    			for (int x=0; x<tilesx ;x++) {
    				char name[64];
    				sprintf(name,"%s%0*d%0*d.bmp",name_out,digits,tilesy-1-pixy/(tilesizey-overlap),digits,x);
    				fp_out1[x] = fopen(name,"wb");
    				printf("Writing %s ...\n",name);
    				if (!fp_out1[x]) throw ERR_WRITE;
    				written = fwrite(&head_out,1,sizeof(head_out),fp_out1[x]);
    				if (written<sizeof(head_out)) throw ERR_WRITE;
    				written = fwrite(&palette,1,palsize_bytes,fp_out1[x]);
    				if (written<palsize_bytes) throw ERR_WRITE;
    			}
  			}
			}
			// SPLIT ONE ROW
			read=fread(&buffer_in,1,rounded_scanline_in,fp_in);
			if (read<rounded_scanline_in) throw ERR_READ;
			for (int x=0; x<tilesx ;x++) {
				bitcopy(buffer_out,buffer_in,0,x*scanline_in_step_bits,scanline_in_bits);
				if (fp_out1[x]) {
					written = fwrite(&buffer_out,1,rounded_scanline_out,fp_out1[x]);
					if (written<rounded_scanline_out) throw ERR_WRITE;
				}
				if (fp_out2[x]) {
  				written = fwrite(&buffer_out,1,rounded_scanline_out,fp_out2[x]);
	  			if (written<rounded_scanline_out) throw ERR_WRITE;
	  		}
			}
      // CLOSE FILES
			if ((pixy+1-overlap)%(tilesizey-overlap)==0 && pixy+1!=overlap) {
  			for (int x=0; x<tilesx ;x++) {
  			  if (overlap) {
  	  			fclose(fp_out2[x]); fp_out2[x]=NULL;
  	  		}
  	  		else {
  	  			fclose(fp_out1[x]); fp_out1[x]=NULL;
  	  		}
		  	}
		  }
		}
		fclose(fp_in); fp_in=NULL;
	}
	catch (char *message) {
		printf("%s\n",message);
		if (fp_in!=NULL) fclose(fp_in);
		for (int i=0; i<MAX_TILES; i++)	{
		  if (fp_out1[i]!=NULL) fclose(fp_out1[i]);
		  if (fp_out2[i]!=NULL) fclose(fp_out2[i]);
		}
		return false;
	}
	return true;
}

const char *usage = "Usage: bmpsplit [-digits N] [-overlap N] in.bmp out_prefix tile_x tile_y\n";

int main(int argc, char *argv[],char *envp[]) {
	int tilesizex = 0, tilesizey = 0, digits = 2, overlap = 0;
	const char *input = NULL, *output = NULL;
	bool ok = false;
	for (int i=1; i<argc; i++) {
		if (strcmp(argv[i],"-digits")==0) {
			if (++i>=argc || !sscanf(argv[i],"%d",&digits)) break;
		}
		else if (strcmp(argv[i],"-overlap")==0) {
			if (++i>=argc || !sscanf(argv[i],"%d",&overlap)) break;
		}
		else {
			if (i+3>=argc) break;
			input  = argv[i];
			output = argv[i+1];
			if (!sscanf(argv[i+2],"%d",&tilesizex)) break;
			if (!sscanf(argv[i+3],"%d",&tilesizey)) break;
			ok = true;
			break;
		}
	}
	if (!ok) { printf(usage); return 1; }
	if (!split_one(input,output,tilesizex,tilesizey,digits,overlap)) return 1;
	return 0;
}
